//  ALL_Relais: Programmierbares Relais
//
//  Unter Verwendung der IoT Brick Library

#include <all_IoT.h>

// *** Globale Typen ***

struct Relais_Parameter
{
//    8 Bytes Parameter-Header:
    int64 parameterID      = 0;                                         // Kennung des Parameter-Blocks "ALrelais" im Klartext
    int64 parameterHash    = 0;                                         // Kennung des Parameter-Blocks "ALrelais" als verschlüsselter Hash
    
//    Start des Parameter-Blocks:
    byte   IP_address0      = 0;                                        // 4 Bytes IP-Adresse
    byte   IP_address1      = 0;
    byte   IP_address2      = 0;
    byte   IP_address3      = 0;
    byte   IP_gateway0      = 0;                                        // 4 Bytes Gateway
    byte   IP_gateway1      = 0;
    byte   IP_gateway2      = 0;
    byte   IP_gateway3      = 0;
    byte   IP_subnet0       = 0;                                        // 4 Bytes Subnetzmaske
    byte   IP_subnet1       = 0;
    byte   IP_subnet2       = 0;
    byte   IP_subnet3       = 0;
    byte   staticIP         = 0;                                        // 0 = Dynamische IP-Adresse, 1 = Statische IP-Adresse
    byte   bootStatus       = 0;                                        // 0 = Relais ausgeschaltet,  1 = Relais eingeschaltet
    byte   listCount        = 0;                                        // Anzahl der Einträge in der Timer-Liste (n)
};

Relais_Parameter Relais_ParameterData;                                  // Parameter Block für das Relais

const byte maxCount    = 50;                                            // Maximale Anzahl Timer & Historie (kleiner gleich 200)

int    counter           = 0;
int    lastMinute        = -1;                                          // Die zuletzt bearbeitete Minute (in 1 Minute nur 1 Aktion)

bool   IntervalActive    = false;                                       // Ist ein Intervall aktuell in Aktion?
int    IntervalStart     = 0;                                           // Das aktuelle Intervall: Start-Minute (0-1439)
int    IntervalEnd       = 0;                                           // Das aktuelle Intervall: Ende-Stunde  (0-1439)
byte   IntervalType      = 0;                                           // Das aktuelle Intervall: Typ (8-11)
byte   Interval1M        = 0;                                           // Das aktuelle Intervall: Intervall 1 in Minuten (1-240)
byte   Interval2M        = 0;                                           // Das aktuelle Intervall: Intervall 2 in Minuten (1-240)
String IntervalInfo      = "";                                          // Das aktuelle Intervall: Klartext der Definition
byte   IntervalAction    = 0;                                           // Das aktuelle Intervall: Aktive Aktion (1-2)
byte   IntervalStatus    = 0;                                           // Das aktuelle Intervall: Aktueller Status des Relais (0-1)
byte   IntervalEndStat   = 0;                                           // Der Status des Relais (0-1) am Ende des Intervalls
int    IntervalNextTime  = 0;                                           // Das aktuelle Intervall: Beginn der nächsten Aktion (0-1439)

byte   Relais_StaticIP   = 0;                                           // 0 = Dynamische IP-Adresse, 1 = Statische IP-Adresse
byte   Relais_BootStatus = 0;                                           // Status des Relais nach einem Neustart (0 oder 1)
byte   Relais_Status     = 0;                                           // Aktueller Status des Relais (0 oder 1)
int    Relais_Delay      = 300;                                         // Zeitspanne für das Schalten des Relais in ms. (300 ms.)

byte   GPIOset           = 14;                                          // GPIO-Nummer für das Setzen des Relais (IoT-Brick: 13)
byte   GPIOreset         = 12;                                          // GPIO-Nummer für das Rücksetzen des Relais (IoT-Brick: 14)
byte   GPIOerr           = 0;                                           // GPIO-Nummer für Fehlermeldung "Not ready.", normale 0

                                                                        // Benötigter Speicherplatz = maxCount * 4 Bytes
byte   listDays          [maxCount];                                    // Wochentage bitweise codiert
byte   listHour          [maxCount];                                    // Start-Stunde (0-23)
byte   listMinute        [maxCount];                                    // Start-Minute (0-59)
byte   listSwitch        [maxCount];                                    // Art des Events (0 = Aus, 1 = Ein, 2 = Inv, 7 = Intervall)
byte   listEndHour       [maxCount];                                    // Ende-Stunde (0-23) - nur bei Intervall
byte   listEndMinute     [maxCount];                                    // Ende-Minute (0-59) - nur bei Intervall
byte   listAction1       [maxCount];                                    // Einschaltdauer in Minuten (1-255) - nur bei Intervall
byte   listAction2       [maxCount];                                    // Ausschaltdauer in Minuten (1-255) - nur bei Intervall

byte   listCount         = 0;                                           // Anzahl der Einträge in der Timer-Liste (n)
byte   listSelected      = 0;                                           // Selektierter Eintrag in der Timer-Liste (0 bis n-1)

String Relais_History    [maxCount];                                    // Die letzten Schaltvorgänge (Historie)
String Relais_ID         = "ESP_UP_Relais " + left(newuniqueid(), 12);  // Eindeutigen Namen erzeugen

void Relais_ShellSortArrays ()
{
  int Anzahl = listCount;
  if (Anzahl > 1)
  {
    int Incr = 1;
    do
    {
      Incr *= 3;
      Incr += 1;
    }
    while (Incr < Anzahl);
    do
    {
      Incr /= 3;
      for (int I = Incr; I < Anzahl; I++)
      {
        int saveDays      = listDays     [I];
        int saveHour      = listHour     [I];
        int saveMinute    = listMinute   [I];
        int saveSwitch    = listSwitch   [I];
        int saveEndHour   = listEndHour  [I];
        int saveEndMinute = listEndMinute[I];
        int saveAction1   = listAction1  [I];
        int saveAction2   = listAction2  [I];
        int TheVal = listHour[I] * 60 + listMinute[I];
        int J = I;
        bool notExit = true;
        while (notExit && (listHour[J - Incr] * 60 + listMinute[J - Incr] > TheVal))
        {
          listDays     [J] = listDays     [J - Incr];
          listHour     [J] = listHour     [J - Incr];
          listMinute   [J] = listMinute   [J - Incr];
          listSwitch   [J] = listSwitch   [J - Incr];
          listEndHour  [J] = listEndHour  [J - Incr];
          listEndMinute[J] = listEndMinute[J - Incr];
          listAction1  [J] = listAction1  [J - Incr];
          listAction2  [J] = listAction2  [J - Incr];
          J -= Incr;
          if (J < Incr)
          {
            notExit = false;
          }
        }
        listDays     [J] = saveDays;
        listHour     [J] = saveHour;
        listMinute   [J] = saveMinute;
        listSwitch   [J] = saveSwitch;
        listEndHour  [J] = saveEndHour;
        listEndMinute[J] = saveEndMinute;
        listAction1  [J] = saveAction1;
        listAction2  [J] = saveAction2;
      }
    }
    while (Incr != 1);
  }
}

void Relais_On ()
{
  Relais_Status = 1;
  EEPROM.put(1023, Relais_Status);                                      // Relais-Status in das EEPROM sichern
  IoT_EEPROMupdate();                                                   // EEPROM aktualisieren
  digitalWrite(GPIOset, HIGH);                                          // GPIO zum Setzen auf 3V schalten
  delay(Relais_Delay);                                                  // min. 300 ms. warten (bis geschaltet wurde)
  digitalWrite(GPIOset, LOW);                                           // GPIO zum Setzen auf 0V schalten 
}

void Relais_Off ()
{
  Relais_Status = 0;
  EEPROM.put(1023, Relais_Status);                                      // Relais-Status in das EEPROM sichern
  IoT_EEPROMupdate();                                                   // EEPROM aktualisieren
  digitalWrite(GPIOreset, HIGH);                                        // GPIO zum Setzen auf 3V schalten
  delay(Relais_Delay);                                                  // min. 300 ms. warten (bis geschaltet wurde)
  digitalWrite(GPIOreset, LOW);                                         // GPIO zum Setzen auf 0V schalten
}

void Relais_Invert ()
{
  if (1 - Relais_Status == 1)
  {
    Relais_On();
  }
  else
  {
    Relais_Off();
  }
}

String Relais_GetTimerStr (byte tDay, byte tHour, byte tMinute, byte tSwitch, byte eHour, byte eMinute, byte i1, byte i2)
{
  String timerString = "";
  if ((tDay & 2) > 0)                                         // Montag
  {
    timerString += "MO ";
  }
  else
  {
    timerString += "   ";
  }
  if ((tDay & 4) > 0)                                         // Dienstag
  {
    timerString += "DI ";
  }
  else
  {
    timerString += "   ";
  }
  if ((tDay & 8) > 0)                                         // Mittwoch
  {
    timerString += "MI ";
  }
  else
  {
    timerString += "   ";
  }
  if ((tDay & 16) > 0)                                        // Donnerstag
  {
    timerString += "DO ";
  }
  else
  {
    timerString += "   ";
  }
  if ((tDay & 32) > 0)                                        // Freitag
  {
    timerString += "FR ";
  }
  else
  {
    timerString += "   ";
  }
  if ((tDay & 64) > 0)                                        // Samstag
  {
    timerString += "SA ";
  }
  else
  {
    timerString += "   ";
  }
  if ((tDay & 1) > 0)                                         // Sonntag
  {
    timerString += "SO ";
  }
  else
  {
    timerString += "   ";
  }
  if ((tHour == 99) || (tMinute == 99) || (eHour == 99) || (eMinute == 99) || (i1 == 0) || (i2 == 0) || (tDay == 0))
  {
    timerString = "(Fehlerhafte Zeitangabe oder kein Wochentag ausgewählt )";
  }
  else
  {
    timerString += "- ";
    timerString += strform(tHour,   48, 2, false);
    timerString += ":";
    timerString += strform(tMinute, 48, 2, false);
    timerString += " - ";
    switch (tSwitch)
    {
      case 0:
        timerString += "Ausschalten";
        break;
      case 1:
        timerString += "Einschalten";
        break;
      case 2:
        timerString += "Invertieren";
        break;
      case 3:
        timerString += "Zufällig schalten";
        break;
      case 8:
        timerString += "Intervall ";
        timerString += strform(i1, 48, 3, false);
        timerString += " Aus & ";
        timerString += strform(i2, 48, 3, false);
        timerString += " Ein, um ";
        timerString += strform(eHour,   48, 2, false);
        timerString += ":";
        timerString += strform(eMinute, 48, 2, false);
        timerString += " Aus";
        break;
      case 9:
        timerString += "Intervall ";
        timerString += strform(i1, 48, 3, false);
        timerString += " Aus & ";
        timerString += strform(i2, 48, 3, false);
        timerString += " Ein, um ";
        timerString += strform(eHour,   48, 2, false);
        timerString += ":";
        timerString += strform(eMinute, 48, 2, false);
        timerString += " Ein";
        break;
      case 10:
        timerString += "Intervall ";
        timerString += strform(i1, 48, 3, false);
        timerString += " Ein & ";
        timerString += strform(i2, 48, 3, false);
        timerString += " Aus, um ";
        timerString += strform(eHour,   48, 2, false);
        timerString += ":";
        timerString += strform(eMinute, 48, 2, false);
        timerString += " Aus";
        break;
      case 11:
        timerString += "Intervall ";
        timerString += strform(i1, 48, 3, false);
        timerString += " Ein & ";
        timerString += strform(i2, 48, 3, false);
        timerString += " Aus, um ";
        timerString += strform(eHour,   48, 2, false);
        timerString += ":";
        timerString += strform(eMinute, 48, 2, false);
        timerString += " Ein";
        break;
      case 12:
        timerString += "Intervall ";
        timerString += strform(i1, 48, 3, false);
        timerString += " Inv ";
        timerString += strform(i2, 48, 3, false);
        timerString += " Inv, um ";
        timerString += strform(eHour,   48, 2, false);
        timerString += ":";
        timerString += strform(eMinute, 48, 2, false);
        timerString += " nicht ändern";
        break;
      default:
        timerString += "(keine Funktion)";
        break;
    }
  }
  return (timerString);
}

void Relais_Trace (String newEvent)
{
  Serial.println(newEvent);
  for (int i = 1; i < maxCount; i++)
  {
    Relais_History[i - 1] = Relais_History[i];
  }
  Relais_History[maxCount - 1] = newEvent;
}

// *** Relais Parameter für das EEPROM des ESP8266 ***

bool Relais_ParameterValidate ()
{
  if ((Relais_ParameterData.parameterID == crc64("ALrelais", 0)) &&
      (Relais_ParameterData.parameterHash == xorshift128plus("ALrelais", 0)))
  {
    return (true);                                                          // Parameter sind gültig
  }
  else
  {
    return (false);                                                         // Ungültige Parameter
  }
}

bool Relais_ParameterApply ()
{
  if (Relais_ParameterValidate())
  {
    if (Relais_ParameterData.staticIP == 1)
    {
      WiFi.localIP()[0] = Relais_ParameterData.IP_address0;                 // 4 Bytes IP-Adresse
      WiFi.localIP()[1] = Relais_ParameterData.IP_address1;
      WiFi.localIP()[2] = Relais_ParameterData.IP_address2;
      WiFi.localIP()[3] = Relais_ParameterData.IP_address3;
      WiFi.gatewayIP()[0] = Relais_ParameterData.IP_gateway0;               // 4 Bytes Gateway
      WiFi.gatewayIP()[1] = Relais_ParameterData.IP_gateway1;
      WiFi.gatewayIP()[2] = Relais_ParameterData.IP_gateway2;
      WiFi.gatewayIP()[3] = Relais_ParameterData.IP_gateway3;
      WiFi.subnetMask()[0] = Relais_ParameterData.IP_subnet0;               // 4 Bytes Subnetzmaske
      WiFi.subnetMask()[1] = Relais_ParameterData.IP_subnet1;
      WiFi.subnetMask()[2] = Relais_ParameterData.IP_subnet2;
      WiFi.subnetMask()[3] = Relais_ParameterData.IP_subnet3;
      IPAddress ip1 = IPAddress(Relais_ParameterData.IP_address0,
                                Relais_ParameterData.IP_address1,
                                Relais_ParameterData.IP_address2,
                                Relais_ParameterData.IP_address3);
      IPAddress ip2 = IPAddress(Relais_ParameterData.IP_gateway0,
                                Relais_ParameterData.IP_gateway1,
                                Relais_ParameterData.IP_gateway2,
                                Relais_ParameterData.IP_gateway3);
      IPAddress ip3 = IPAddress(Relais_ParameterData.IP_subnet0,
                                Relais_ParameterData.IP_subnet1,
                                Relais_ParameterData.IP_subnet2,
                                Relais_ParameterData.IP_subnet3);
      WiFi.config(ip1, ip2, ip3);                                           // Parameter: IP, Gateway, Subnet, DNS
    }
    Relais_StaticIP = Relais_ParameterData.staticIP;                        // 0 = Dynamische IP-Adresse, 1 = Statische IP-Adresse
    Relais_BootStatus = Relais_ParameterData.bootStatus;                    // Status des Relais nach einem Neustart (0 oder 1)
    listCount = Relais_ParameterData.listCount;                             // Anzahl der Einträge in der Timer-Liste (n)
    return (true);                                                          // Parameter-Block ist gültig und wurde verwendet
  }
  else
  {
    return (false);                                                         // Parameter-Block ist ungültig und wurde nicht verwendet
  }
}

void Relais_ParameterCreate ()
{
//   32 Bytes Parameter-Header:
  Relais_ParameterData.parameterID      = crc64("ALrelais", 0);             // Kennung des Parameter-Blocks "ALrelais" im Klartext
  Relais_ParameterData.parameterHash    = xorshift128plus("ALrelais", 0);   // Kennung des Parameter-Blocks "ALrelais" als verschlüsselter Hash
    
//   Start des Parameter-Blocks:
  Relais_ParameterData.IP_address0      = WiFi.localIP()[0];                // 4 Bytes IP-Adresse
  Relais_ParameterData.IP_address1      = WiFi.localIP()[1];
  Relais_ParameterData.IP_address2      = WiFi.localIP()[2];
  Relais_ParameterData.IP_address3      = WiFi.localIP()[3];
  Relais_ParameterData.IP_gateway0      = WiFi.gatewayIP()[0];              // 4 Bytes Gateway
  Relais_ParameterData.IP_gateway1      = WiFi.gatewayIP()[1];
  Relais_ParameterData.IP_gateway2      = WiFi.gatewayIP()[2];
  Relais_ParameterData.IP_gateway3      = WiFi.gatewayIP()[3];
  Relais_ParameterData.IP_subnet0       = WiFi.subnetMask()[0];             // 4 Bytes Subnetzmaske
  Relais_ParameterData.IP_subnet1       = WiFi.subnetMask()[1];
  Relais_ParameterData.IP_subnet2       = WiFi.subnetMask()[2];
  Relais_ParameterData.IP_subnet3       = WiFi.subnetMask()[3];
  Relais_ParameterData.staticIP         = Relais_StaticIP;                  // 0 = Dynamische IP-Adresse, 1 = Statische IP-Adresse
  Relais_ParameterData.bootStatus       = Relais_BootStatus;                // Status des Relais nach einem Neustart (0 oder 1)
  Relais_ParameterData.listCount        = listCount;                        // Anzahl der Einträge in der Timer-Liste (n)
}

void Relais_ParameterNew ()
{
//   32 Bytes Parameter-Header:
  Relais_ParameterData.parameterID      = crc64("ALrelais", 0);             // Kennung des Parameter-Blocks "ALrelais" im Klartext
  Relais_ParameterData.parameterHash    = xorshift128plus("ALrelais", 0);   // Kennung des Parameter-Blocks "ALrelais" als verschlüsselter Hash
    
//   Start des Parameter-Blocks:
  Relais_ParameterData.IP_address0      = 0;                                // 4 Bytes IP-Adresse
  Relais_ParameterData.IP_address1      = 0;
  Relais_ParameterData.IP_address2      = 0;
  Relais_ParameterData.IP_address3      = 0;
  Relais_ParameterData.IP_gateway0      = 0;                                // 4 Bytes Gateway
  Relais_ParameterData.IP_gateway1      = 0;
  Relais_ParameterData.IP_gateway2      = 0;
  Relais_ParameterData.IP_gateway3      = 0;
  Relais_ParameterData.IP_subnet0       = 0;                                // 4 Bytes Subnetzmaske
  Relais_ParameterData.IP_subnet1       = 0;
  Relais_ParameterData.IP_subnet2       = 0;
  Relais_ParameterData.IP_subnet3       = 0;
  Relais_ParameterData.staticIP         = 0;                                // 0 = Dynamische IP-Adresse, 1 = Statische IP-Adresse
  Relais_ParameterData.bootStatus       = 0;                                // Status des Relais nach einem Neustart (0 oder 1)
  Relais_ParameterData.listCount        = 0;                                // Anzahl der Einträge in der Timer-Liste (n)
}

bool Relais_ParameterLoad ()
{
  EEPROM.get(0x000, Relais_ParameterData);                                  // Parameter aus dem EEPROM laden (0x3FF = aktueller Status)
  if (Relais_ParameterValidate())                                           // Die geladenen Daten validieren
  {
    EEPROM.get(0x400, listDays);                                            // Timer-Wochentage in das EEPROM sichern
    EEPROM.get(0x600, listHour);                                            // Timer-Start-Stunden aus dem EEPROM laden
    Relais_ID = IoT_EEPROMgetString(0x500);                                 // Geräte-Namen aus dem EEPROM laden
    EEPROM.get(0x700, listEndHour);                                         // Timer-Ende-Stunden aus dem EEPROM laden
    EEPROM.get(0x800, listMinute);                                          // Timer-Start-Minuten aus dem EEPROM laden
    EEPROM.get(0x900, listEndMinute);                                       // Timer-Ende-Minuten aus dem EEPROM laden
    EEPROM.get(0xA00, listSwitch);                                          // Timer-Aktionen aus dem EEPROM laden
    EEPROM.get(0xB00, listAction1);                                         // Timer-Einschaltzeit für Intervall aus dem EEPROM laden
    EEPROM.get(0xC00, listAction2);                                         // Timer-Ausschaltzeit für Intervall aus dem EEPROM laden
    if (Relais_ID == "")
    {
      Relais_ID = "ESP_UP_Relais " + left(newuniqueid(), 12);
    }
    return (true);                                                          // Parameter sind gültig und wurden erfolgreich geladen
  }
  else
  {
    return (false);                                                         // Ungültige Parameter
  }
}

void Relais_ParameterSave ()
{
  if (Relais_ParameterValidate())
  {
    Relais_ID = left(Relais_ID, 63);                                        // maximale Länge = 63 Zeichen
    if (Relais_ID == "")
    {
      Relais_ID = "ESP_UP_Relais " + left(newuniqueid(), 12);
    }
    EEPROM.put(0x000, Relais_ParameterData);                                // Parameter in das EEPROM sichern (0x3FF = aktueller Status)
    EEPROM.put(0x400, listDays);                                            // Timer-Wochentage in das EEPROM sichern
    IoT_EEPROMputString(0x500, Relais_ID);
    EEPROM.put(0x600, listHour);                                            // Timer-Start-Stunden in das EEPROM sichern
    EEPROM.put(0x700, listEndHour);                                         // Timer-Ende-Stunden in das EEPROM sichern                                   
    EEPROM.put(0x800, listMinute);                                          // Timer-Start-Minuten in das EEPROM sichern
    EEPROM.put(0x900, listEndMinute);                                       // Timer-Ende-Minuten in das EEPROM sichern
    EEPROM.put(0xA00, listSwitch);                                          // Timer-Aktionen in das EEPROM sichern
    EEPROM.put(0xB00, listAction1);                                         // Timer-Einschaltzeit für Intervall in das EEPROM sichern
    EEPROM.put(0xC00, listAction2);                                         // Timer-Ausschaltzeit für Intervall in das EEPROM sichern
  }
}

void Relais_SetInterval (int i)
{
  int h = hour();
  int m = minute();
  int minuteOfDay = h * 60 + m;
  IntervalActive   = true;                                                  // Aktives Intervall ist in Benutzung
  IntervalStart    = listHour[i] * 60 + listMinute[i];                      // Das aktuelle Intervall: Start in der Minute des Tages (0-1439)
  IntervalEnd      = listEndHour[i] * 60 + listEndMinute[i];                // Das aktuelle Intervall: Ende in der Minute des Tages (0-1439)
  IntervalType     = listSwitch[i];                                         // Das aktuelle Intervall: Typ (8-11), 255 = Kein Intervall
  Interval1M       = listAction1[i];                                        // Das aktuelle Intervall: Intervall 1 in Minuten (1-240)
  Interval2M       = listAction2[i];                                        // Das aktuelle Intervall: Intervall 2 in Minuten (1-240)
  IntervalAction   = 1;                                                     // Die gerade aktive Aktion des Intervalls (1-2)
  IntervalInfo     = Relais_GetTimerStr(listDays[i], listHour[i], listMinute[i], IntervalType, listEndHour[i], listEndMinute[i], Interval1M, Interval2M);
  IntervalNextTime = minuteOfDay + Interval1M;                              // Beginn der nächsten Aktion des Intervalls (0-1439)
  if (IntervalNextTime > 1439)                                              // Ende am nächsten Tag?
  {
    IntervalNextTime = 1439;
  }
  if (IntervalNextTime > IntervalEnd)                                       // Ende hinter der letzten Schaltung?
  {
    IntervalNextTime = IntervalEnd;
  }
}

bool Relais_TestWeekDay (int i)
{
  bool action = false;
  switch (weekday())
  {
    case 1: // Sonntag
      action = (listDays[i] & 1) > 0;
      break;
    case 2: // Montag
      action = (listDays[i] & 2) > 0;
      break;
    case 3: // Dienstag
      action = (listDays[i] & 4) > 0;
      break;
    case 4: // Mittwoch
      action = (listDays[i] & 8) > 0;
      break;
    case 5: // Donnerstag
      action = (listDays[i] & 16) > 0;
      break;
    case 6: // Freitag
      action = (listDays[i] & 32) > 0;
      break;
    case 7: // Samstag
      action = (listDays[i] & 64) > 0;
      break;
    default:
      break;
  }
  return (action);
}

void Relais_StartAction (int i)
{
  String newEvent = IoT_NTPtime() + ": ";
  byte rb = (rndbyte() & 1);                              // 1 Bit Zufallszahl
  switch (listSwitch[i])
  {
    case 0: // Ausschalten
      newEvent += "Relais ausgeschaltet.";
      Relais_Off();
      break;
    case 1: // Einschalten
      newEvent += "Relais eingeschaltet.";
      Relais_On();
      break;
    case 2: // Invertieren
      Relais_Invert();
      if (Relais_Status == 0)
      {
        newEvent += "Relais invertiert (ausgeschaltet).";
      }
      else if (Relais_Status == 1)
      {
        newEvent += "Relais invertiert (eingeschaltet).";
      }
      break;
    case 3: // Zufällig schalten
      if (rb == 0)
      {
        newEvent += "Relais zufällig ausgeschaltet.";
        Relais_Off();
      }
      else if (rb == 1)
      {
        newEvent += "Relais zufällig eingeschaltet.";
        Relais_On();
      }
      break;
    case 8: // Intervall aus-ein -> aus
      Relais_SetInterval(i);
      IntervalStatus   = 0;                                       // Der aktuelle Status des Relais (0-1)
      IntervalEndStat  = 0;                                       // Der Status des Relais (0-1) am Ende des Intervalls
      newEvent += "Relais beim Intervall-Start ausgeschaltet.";
      Relais_Off();
      break;
    case 9: // Intervall aus-ein -> ein
      Relais_SetInterval(i);
      IntervalStatus   = 0;                                       // Der aktuelle Status des Relais (0-1)
      IntervalEndStat  = 1;                                       // Der Status des Relais (0-1) am Ende des Intervalls
      newEvent += "Relais beim Intervall-Start ausgeschaltet.";
      Relais_Off();
      break;
    case 10: // Intervall ein-aus -> aus
      Relais_SetInterval(i);
      IntervalStatus   = 1;                                       // Der aktuelle Status des Relais (0-1)
      IntervalEndStat  = 0;                                       // Der Status des Relais (0-1) am Ende des Intervalls
      newEvent += "Relais beim Intervall-Start eingeschaltet.";
      Relais_On();
      break;
    case 11: // Intervall ein-aus -> ein
      Relais_SetInterval(i);
      IntervalStatus   = 1;                                       // Der aktuelle Status des Relais (0-1)
      IntervalEndStat  = 1;                                       // Der Status des Relais (0-1) am Ende des Intervalls
      newEvent += "Relais beim Intervall-Start eingeschaltet.";
      Relais_On();
      break;
    case 12: // Intervall Invertieren
      Relais_SetInterval(i);
      Relais_Invert();
      IntervalStatus  = Relais_Status;                            // Der aktuelle Status des Relais (0-1)
      IntervalEndStat = 2;                                        // Am Ende des Intervalls nichts verändern
      if (Relais_Status == 0)
      {
        newEvent += "Relais beim Intervall-Start invertiert (ausgeschaltet).";
      }
      else if (Relais_Status == 1)
      {
        newEvent += "Relais beim Intervall-Start invertiert (eingeschaltet).";
      }
      break;
    default:
      newEvent += "Illegale Aktion ignoriert.";
      break;
  }
  Relais_Trace(newEvent);
}
            
void setup()
{
  IoT_Init(true);                                                   // true = IoT-Brick, false = Relais-Platine
  t_TerminalInit();                                                 // Terminal initialisieren (incl. Terminal-Bildschirm löschen)
  IoT_DisplayClear(24);                                             // OLED Display löschen (24 Punkt Schrift)
  IoT_DisplayAlignText(TEXT_ALIGN_CENTER);                          // Zentrierte Darstellung des Textes
  IoT_DisplayDrawText(64, 5, "Reset");
  IoT_DisplayDrawText(64, 34, "Konfig.?");
  IoT_DisplayUpdate();                                              // OLED-Anzeige aktualisieren
  IoT_WaitKeypress(2000, false);                                    // 2 s warten, in der Zeit kann mit Taster die Konfig zurückgesetzt werden
  if (IoT_Keypress())                                               // Wenn Taster gedrückt wird, Konfiguration im EEPROM löschen
  {
    IoT_EEPROMclear();                                              // Alle Parameter aus dem EEPROM löschen
    IoT_EEPROMupdate();                                             // EEPROM aktualisieren
    Serial.print("Relais Parameter aus EEPROM gelöscht. ");
    Serial.println("Relais wird neu gestartet...");
    delay(100);                                                     // 100 ms. warten
    IoT_DisplayClear(24);                                           // OLED Display löschen (24 Punkt Schrift)
    IoT_DisplayUpdate();                                            // OLED-Anzeige aktualisieren
    IoT_WaitNoKeypress();                                           // Warten bis der Taster losgelassen wurde
    t_Restart();                                                    // Neustart durchführen
  }
  if (Relais_ParameterLoad())                                       // Parameter-Block aus EEPROM laden
  {
    Serial.println("Relais Parameter aus EEPROM geladen.");
    Relais_ParameterApply();                                        // Parameter-Block in die aktuelle Konfiguration übernehmen
  }
  else
  {
    Serial.println("Relais Parameter im EEPROM nicht vorhanden.");
  }
  IoT_DisplayClear(24);                                             // OLED Display löschen (24 Punkt Schrift)
  IoT_DisplayAlignText(TEXT_ALIGN_CENTER);                          // Zentrierte Darstellung des Textes
  IoT_DisplayDrawText(64, 5, "WLAN");
  IoT_DisplayDrawText(64, 34, "suchen...");
  IoT_DisplayUpdate();                                              // OLED-Anzeige aktualisieren
  pinMode(GPIOset, OUTPUT);                                         // GPIO zum Setzen auf Ausgabe schalten
  pinMode(GPIOreset, OUTPUT);                                       // GPIO zum Zurücksetzen auf Ausgabe schalten
  pinMode(GPIOerr, OUTPUT);                                         // GPIO für Fehlerausgabe auf Ausgabe schalten
  digitalWrite(GPIOset, LOW);                                       // GPIO zum Setzen auf 0V schalten
  digitalWrite(GPIOreset, LOW);                                     // GPIO zum Zurücksetzen auf 0V schalten
  if (Relais_BootStatus == 0)
  {
    Relais_Off();
  }
  else if (Relais_BootStatus == 1)
  {
    Relais_On();
  }
  else if (Relais_BootStatus == 2)
  {
    EEPROM.get(1023, Relais_Status);                                // Letzten Status des Relais abfragen
    if (Relais_Status == 0)
    {
      Relais_Off();
    }
    else if (Relais_Status == 1)
    {
      Relais_On();
    }
  }
  digitalWrite(GPIOerr, LOW);                                       // Fehler-GPIO einschalten
  IoT_WLANautoConnect(Relais_ID, true);                             // Verbindung, ggf. gespeicherte, über WLAN herstellen
  IoT_NTPinit();                                                    // NTP-Timeserver initialisieren
  IoT_DisplayClear(24);                                             // OLED Display löschen (24 Punkt Schrift)
  IoT_DisplayAlignText(TEXT_ALIGN_CENTER);                          // Zentrierte Darstellung des Textes
  IoT_DisplayDrawText(64, 5, "NTP-Server");
  IoT_DisplayDrawText(64, 34, "suchen...");
  IoT_DisplayUpdate();                                              // OLED-Anzeige aktualisieren
  while (not(IoT_NTPvalid()))                                       // Prüfen, ob eine gültige Uhrzeit aus dem Internet empfangen wurde
  {
    IoT_Idle();
    delay(100);                                                     // 100 ms. warten
  }
  Serial.println("Connected to NTP Time-Server");
  if (Relais_BootStatus == 0)
  {
    Relais_Trace(IoT_NTPtime() + ": Relais beim Neustart ausgeschaltet.");
  }
  else if (Relais_BootStatus == 1)
  {
    Relais_Trace(IoT_NTPtime() + ": Relais beim Neustart eingeschaltet.");
  }
  else if (Relais_BootStatus == 2)
  {
    bool actionDone = false;
    if (listCount > 0)                                              // Nach einem aktuell laufenden Intervall suchen
    {
      int lastElement = listCount - 1;
      for (int i = 0; i <= lastElement; i++)                        // Alle Timer-Einträge durchsuchen
      {
        if (Relais_TestWeekDay(i))                                  // Richtiger Wochentag
        {
          int h = hour();
          int m = minute();
          int sH = listHour[i];
          int sM = listMinute[i];
          if ((h > sH) || ((h == sH) && (m >= sM)))                 // Richtige Zeit oder später
          {
            int eH = listEndHour[i];
            int eM = listEndMinute[i];
            if ((h < eH) || ((h == eH) && (m <= eM)))               // Am Ende oder vor dem Ende des Intervalls
            {
              lastMinute = -1;
              Relais_Trace(IoT_NTPtime() + ": Relais beim Neustart aktives Intervall wiederherstellen.");
              Relais_StartAction(i);
              lastMinute = m;
              actionDone = true;
            }                                                       // In dieser Minute wurde noch nicht geschaltet
          }                                                         // Richtige Zeit
        }                                                           // Richtiger Wochentag
      }                                                             // for-Schleife
    }
    if (not(actionDone))                                            // Nur wenn keine Aktion gefunden wurde
    {
      if (Relais_Status == 0)
      {
        Relais_Trace(IoT_NTPtime() + ": Relais beim Neustart den letzten Zustand hergestellt (ausgeschaltet).");
      }
      else if (Relais_Status == 1)
      {
        Relais_Trace(IoT_NTPtime() + ": Relais beim Neustart den letzten Zustand hergestellt (eingeschaltet).");
      }
    }
  }
  IoT_WebServer.on("/", handleRoot);                                // Browser-Handler (siehe unten) für das Stammverzeichnis
  IoT_WebServer.on("/Reload", handleRoot);                          // Handler für Reload
  IoT_WebServer.on("/ButtonOn",       handleButtonOn);              // Handler für Ein-Schalter
  IoT_WebServer.on("/ButtonOff",      handleButtonOff);             // Handler für Aus-Schalter
  IoT_WebServer.on("/ButtonInvert",   handleButtonInvert);          // Handler für Invertierte Schalter-Position
  IoT_WebServer.on("/SelectPrev",     handleSelectPrev);            // Handler für Nächsten Eintrag auswählen
  IoT_WebServer.on("/SelectNext",     handleSelectNext);            // Handler für Vorherigen Eintrag auswählen
  IoT_WebServer.on("/Abbrechen",      handleAbbrechen);             // Handler für Abbrechen
  IoT_WebServer.on("/Restart",        handleRestart);               // Handler für Neustart
  IoT_WebServer.on("/DelHistory",     handleDelHistory);            // Handler für Historie löschen
  IoT_WebServer.on("/AddTimer",       handleAddTimer);              // Handler für Aktion hinzufügen
  IoT_WebServer.on("/AddTimerOK",     handleAddTimerOK);            // Handler für Aktion hinzufügen OK
  IoT_WebServer.on("/DeleteTimer",    handleDeleteTimer);           // Handler für Aktion löschen
  IoT_WebServer.on("/DeleteAllTimer", handleDeleteAllTimer);        // Handler für Alle Aktionen löschen
  IoT_WebServer.on("/StopInterval",   handleStopInterval);          // Handler für Intervall abbrechen
  IoT_WebServer.on("/Preferences",    handlePreferences);           // Handler für Netzwerkeinstellungen
  IoT_WebServer.on("/NameChange",     handleNameChange);            // Handler für Namen ändern
  IoT_WebServer.on("/NetworkChange",  handleNetworkChange);         // Handler für Netzwerkeinstellungen ändern
  IoT_WebServer.on("/PreferencesOK",  handlePreferencesOK);         // Handler für Netzwerkeinstellungen
  IoT_WebServer.on("/DynamicIP",      handleDynamicIP);             // Handler für dynamische IP-Adresse
  IoT_WebServer.on("/SavePrefs",      handleSavePrefs);             // Handler für Einstellungen dauerhaft speichern
  IoT_WebServer.on("/StandardPrefs",  handleStandardPrefs);         // Handler für Standardeinstellungen widerherstellen
  IoT_WebServer.begin();                                            // ab jetzt "horcht" der Server auf HTTP-Anfragen
  IoT_WebServer.handleClient();                                     // Bediene die http Anfragen
  Serial.println("HTTP server started");
  digitalWrite(GPIOerr, HIGH);                                      // Fehler-GPIO ausschalten
  pinMode(GPIOerr, INPUT);                                          // GPIO für Fehlerausgabe auf Eingabe schalten
}

void loop()
{
  IoT_Idle();
  IoT_WebServer.handleClient();                                     // Bediene die http Anfragen
  int h = hour();
  int m = minute();
  int minuteOfDay = h * 60 + m;
  if (m != lastMinute)
  {
    lastMinute = -1;
  }

  if (IoT_Keypress())                                               // Wenn Taster gedrückt wird, Konfiguration zeigen
  {
    IoT_DisplayClear(10);                                           // OLED Display löschen (10 Punkt Schrift)
    String state = IoT_WLANstatus();                                // Verbindungsstatus
    if (IoT_WLANinited)                                             // Falls WLAN Verbindung erfolgreich, Status & IP im OLED Display anzeigen
    {
      IoT_DisplayDrawText(5, 0, "WLAN: " + IoT_WLANssid());         // WLAN Netzwerk Name
      if (state == "Connected")                                     // Wenn erfolgreich verbunden, dann Signalstärke hinzufügen
      {
        state = state + " (" + str(IoT_WLANrssi()) + "dBm)";        // Signalstärke
      }
      IoT_DisplayDrawText(5, 10, state);                            // Verbindungsstatus & Signalstärke
      IoT_DisplayDrawText(5, 20, "IP: " + IoT_WLANaddress(true));   // IP Adresse
      IoT_DisplayDrawText(5, 30, "GW: " + IoT_WLANgateway(true));   // IP Gateway
      IoT_DisplayDrawText(5, 40, "NT: " + IoT_WLANsubnet(true));    // IP Subnetzmaske
    }
    else                                                            // Falls WLAN Verbindung gescheitert ist, Status im OLED Display anzeigen
    {
      IoT_DisplayDrawText(5, 10, state); 
    }
    IoT_DisplayUpdate();                                            // OLED-Anzeige aktualisieren
    
    String newEvent = IoT_NTPtime() + ": ";
    Relais_Invert();                                                // Beim Drücken des Knopfes Relais invertieren
    if (Relais_Status == 0)
    {
      newEvent += "Relais mit Geräte-Button invertiert (ausgeschaltet).";
    }
    else if (Relais_Status == 1)
    {
      newEvent += "Relais mit Geräte-Button invertiert (eingeschaltet).";
    }
    Relais_Trace(newEvent);
  }
  else
  {
    IoT_DisplayClear(24);                                           // OLED Display löschen (24 Punkt Schrift)
    IoT_DisplayAlignText(TEXT_ALIGN_CENTER);                        // Zentrierte Darstellung des Textes
    IoT_DisplayDrawText(64, 5, "Relais");
    IoT_DisplayDrawText(64, 34, "bereit");
    IoT_Idle();
    if (counter % 100 == 0)                                         // Infos seriell nicht zu oft ausgeben (ca. alle 10 Sekunden)
    {
      if (IoT_NTPeventAvail)                                        // Zeitevent triggern
      {
        // IoT_NTPprintEvent(IoT_NTPcurrentEvent);
        IoT_NTPeventAvail = false;
      }
    }
    IoT_DisplayUpdate();                                            // OLED-Anzeige aktualisieren
  }

  if (IntervalActive)                                               // Ist ein Intervall gerade aktiv?
  {
    if (m != lastMinute)                                            // In dieser Minute wurde noch nicht geschaltet
    {
      if (minuteOfDay == IntervalNextTime)                          // Nächste Intervall-Aktion soll beginnen
      {
        String newEvent = IoT_NTPtime() + ": ";
        if (minuteOfDay >= IntervalEnd)                             // Ende des Intervalls erreicht oder überschritten?
        {
          IntervalActive = false;
          if (IntervalEndStat == 0)
          {
            if (minuteOfDay > IntervalEnd)
            {
              newEvent += "Relais nach Intervall-Ende ausgeschaltet.";
            }
            else
            {
              newEvent += "Relais beim Intervall-Ende ausgeschaltet.";
            }
            Relais_Off();
          }
          else if (IntervalEndStat == 1)
          {
            if (minuteOfDay > IntervalEnd)
            {
              newEvent += "Relais nach Intervall-Ende eingeschaltet.";
            }
            else
            {
              newEvent += "Relais beim Intervall-Ende eingeschaltet.";
            }
            Relais_On();
          }
          else if (IntervalEndStat == 2)
          {
            if (minuteOfDay > IntervalEnd)
            {
              newEvent += "Relais nach Intervall-Ende nicht ändern ";
            }
            else
            {
              newEvent += "Relais beim Intervall-Ende nicht ändern ";
            }
            if (Relais_Status == 0)
            {
              newEvent += "(ausgeschaltet).";
            }
            else if (Relais_Status == 1)
            {
              newEvent += "(eingeschaltet).";
            }
          }
        }
        else
        {
          if (IntervalStatus == 1)                                  // Der aktuelle Status des Relais (0-1)
          {
            IntervalStatus = 0;
            newEvent += "Relais innerhalb des Intervalls ausgeschaltet.";
            Relais_Off();
          }
          else if (IntervalStatus == 0)
          {
            IntervalStatus = 1;
            newEvent += "Relais innerhalb des Intervalls eingeschaltet.";
            Relais_On();
          }
          else
          {
            newEvent += "Illegaler Status (" + str(IntervalStatus) + ") innerhalb des Intervalls.";
            IntervalStatus = 0;
          }
          if (IntervalAction == 1)                                  // Die gerade aktive Aktion des Intervalls (1-2)
          {
            IntervalAction = 2;
            IntervalNextTime += Interval2M;                         // Beginn der nächsten Aktion des Intervalls (0-1439)
          }
          else if (IntervalAction == 2)
          {
            IntervalAction = 1;
            IntervalNextTime += Interval1M;                         // Beginn der nächsten Aktion des Intervalls (0-1439)
          }
          else
          {
            newEvent += "Illegale Aktion (" + str(IntervalAction) + ") innerhalb des Intervalls.";
            IntervalAction = 1;
          }
        }
        if (IntervalNextTime > IntervalEnd)
        {
          IntervalNextTime = IntervalEnd;
        }
        Relais_Trace(newEvent);
        lastMinute = m;
      } 
    }
  }
  else if (listCount > 0)                                           // Timer-Einträge vorhanden & kein aktives Intervall
  {
    int lastElement = listCount - 1;
    for (int i = 0; i <= lastElement; i++)                          // Alle Timer-Einträge durchsuchen
    {
      if (Relais_TestWeekDay(i))                                    // Richtiger Wochentag
      {
        if ((h == listHour[i]) && (m == listMinute[i]))             // Richtige Zeit
        {
          if (m != lastMinute)                                      // In dieser Minute wurde noch nicht geschaltet
          {
            Relais_StartAction(i);
            lastMinute = m;
          }                                                         // In dieser Minute wurde noch nicht geschaltet
        }                                                           // Richtige Zeit
      }                                                             // Richtiger Wochentag
    }                                                               // for-Schleife
  }
  delay(100);                                                       // 100 ms. warten
  counter++;
}

String StandardHeader ()
{
  String ColorHexVal = "#000000";
  String title = "ESP_UP_Relais - IP: " + IoT_WLANaddress(false);
  String time_web = IoT_NTPtime();
  String date_web = IoT_NTPdate();
  String weekday_web = dayname(weekday() - 1);
  String content = "<!DOCTYPE html>";
  String stat = "";
  if (Relais_Status == 1)
  {
    stat = "(eingeschaltet)";
  }
  else
  {
    stat = "(ausgeschaltet)";
  }
  content += "<html>";
  content += "<head>";
  content += "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">";
  content += "<title>" + title + "</title>";
  content += "<style> body { background-color: #FFFFFF; font-family: Menlo, Monaco, Courier, monospace; Color: " + ColorHexVal + "; } </style>";
  content += "</head>";
  content += "<body>";
  content += "<img src='http://newsletter.allnet.de/mailing/newsletteruploads/pm/default/logos/own/logo-allnet.png' alt='Allnet Logo'>";
  content += "<h1> " + title + " </h1>";
  content += "<h2> Elektronisches Relais " + stat + "</h2>";
  content += "<h2>" + Relais_ID + "</h2>";
  content += "<h2>" + weekday_web + " den " + date_web + " um " + time_web + "</h2>";
  return (content);
}

void handleRoot ()
{
  String content = StandardHeader();
  content += "<p>" + cleanhtml(fillcenter(" Manuell schalten ", "—", 85)) + "</p>";
  content +=  IoT_WebFormActionButton ("ButtonOn", "Relais manuell einschalten");
  content +=  IoT_WebFormActionButton ("ButtonOff", "Relais manuell ausschalten");
  content +=  IoT_WebFormActionButton ("ButtonInvert", "Relais manuell invertieren");
  content += "</p>";
  content += "<p>" + cleanhtml(fillcenter(" Timer ", "—", 85)) + "</p>";
  if (listCount > 0) // Anzahl Timer-Einträge
  {
    content += "<p>";
    content += IoT_WebFormActionButton("SelectPrev", "Vorigen Eintrag auswählen");
    content += IoT_WebFormActionButton("SelectNext", "Nächsten Eintrag auswählen");
    content += "<br>";
    for (int i = 0; i < listCount; i++)
    {
      if (i == listSelected)
      {
        content += "> ";
      }
      else
      {
        content += cleanhtml("  ");
      }
      content += cleanhtml(Relais_GetTimerStr (listDays[i], listHour[i], listMinute[i], listSwitch[i], listEndHour[i], listEndMinute[i], listAction1[i], listAction2[i]));
      if (i == listSelected)
      {
        content += " <";
      }
      content += "<br>";
    }
    content += "</p>";
  }
  content +=  IoT_WebFormActionButton ("AddTimer", "Aktion hinzufügen (" + str(listCount) + " von maximal " + str(maxCount) + " benutzt)");
  if (listCount > 0) // Anzahl Timer-Einträge
  {
    content += "<br>";
    content +=  IoT_WebFormActionButton ("DeleteTimer", "Aktion löschen");
    content +=  IoT_WebFormActionButton ("DeleteAllTimer", "Alle Aktionen löschen");
  }
  if (IntervalActive)
  {
    content += "<p>" + cleanhtml(fillcenter(" Aktives Intervall ", "—", 85)) + "</p>";
    content += cleanhtml("  ");
    content += IntervalInfo;
    content += "<br> <br>";
    content +=  IoT_WebFormActionButton ("StopInterval", "Intervall Abbrechen");
  }
  content += "<p>" + cleanhtml(fillcenter(" Historie ", "—", 85)) + "</p>";
  for (int i = 0; i < maxCount; i++)
  {
    if (Relais_History[i].length() > 0)
    {
      content += cleanhtml("  ");
      content += Relais_History[i];
      content += "<br>";
    }
  }
  content += "<br>";
  content +=  IoT_WebFormActionButton("DelHistory", "Historie löschen");
  content +=  IoT_WebFormActionButton("Preferences", "Einstellungen");
  content +=  IoT_WebFormActionButton("Reload", "Seite neu laden");
  content +=  IoT_WebFormActionButton("Restart", "Relais neu starten");
  content += "</body>";
  content += "</html>";
  IoT_WebServer.send(200, "text/html", content);   // HTTP-Statuscode 200 = OK (Anfrage wurde erfolgreich bearbeitet)
}

void handleButtonOn ()
{
  Relais_On();
  Relais_Trace(IoT_NTPtime() + ": Relais manuell eingeschaltet.");
  handleRoot();
}

void handleButtonOff ()
{
  Relais_Off();
  Relais_Trace(IoT_NTPtime() + ": Relais manuell ausgeschaltet.");
  handleRoot();
}

void handleButtonInvert ()
{
  Relais_Invert();
  if (Relais_Status == 0)
  {
    Relais_Trace(IoT_NTPtime() + ": Relais manuell invertiert (ausgeschaltet).");
  }
  else if (Relais_Status == 1)
  {
    Relais_Trace(IoT_NTPtime() + ": Relais manuell invertiert (eingeschaltet).");
  }
  handleRoot();
}

void handleSelectPrev ()
{
  listSelected -= 1;
  if (listSelected < 0)
  {
    listSelected = 0;
  }
  handleRoot();
}

void handleSelectNext ()
{
  listSelected += 1;
  if (listSelected >= listCount)
  {
    listSelected = listCount - 1;
  }
  if (listSelected < 0)
  {
    listSelected = 0;
  }
  handleRoot();
}

void handleAbbrechen ()
{
  handleRoot();
}

void handleRestart ()                                               // Neustart
{
  handleRoot();                                                     // Einstellungen anzeigen
  IoT_Idle();
  t_Restart();
}

void handleDeleteTimer ()
{
  int lastElement = listCount - 1;
  if (listCount > 0)
  {
    if (listSelected == lastElement)
    {
      listCount -= 1;
      listSelected -= 1;
      if (listSelected < 0)
      {
        listSelected = 0;
      }
    }
    else
    {
      for (int i = listSelected; i <= lastElement; i++)
      {
        listDays  [i] = listDays  [i + 1];
        listHour  [i] = listHour  [i + 1];
        listMinute[i] = listMinute[i + 1];
        listSwitch[i] = listSwitch[i + 1];
      }
      listCount -= 1;
    }
  }
  handleRoot();
}

void handleDelHistory ()
{
  for (int i = 0; i < maxCount; i++)
  {
    Relais_History[i] = "";
  }
  Relais_Trace(IoT_NTPtime() + ": Historie manuell gelöscht.");
  handleRoot();
}

void handleDeleteAllTimer ()
{
  listCount = 0;
  handleRoot();
}

void handleStopInterval ()
{
  IntervalActive = false;
  handleRoot();
}

void handleAddTimer ()
{
  if (listCount < maxCount)
  {
    String content = StandardHeader();
    content += IoT_WebFormOpen("AddTimerOK");
    content += "<p>" + cleanhtml(fillcenter(" Timer ", "—", 85)) + "</p>";
    content += IoT_WebCheckBox(" Montag  ",     "day1", "chkd", 1);
    content += IoT_WebCheckBox(" Dienstag  ",   "day2", "chkd", 1);
    content += IoT_WebCheckBox(" Mittwoch  ",   "day3", "chkd", 1);
    content += IoT_WebCheckBox(" Donnerstag  ", "day4", "chkd", 1);
    content += IoT_WebCheckBox(" Freitag  ",    "day5", "chkd", 1);
    content += IoT_WebCheckBox(" Samstag  ",    "day6", "chkd", 1);
    content += IoT_WebCheckBox(" Sonntag  ",    "day7", "chkd", 1);
    content += "<br><br>";
    content += IoT_WebInput("Zeitpunkt: Stunde: ", "actH", "00", 3);
    content += IoT_WebInput("  Minute: ", "actM", "00", 3);
    content += "<p>" + cleanhtml(fillcenter(" Aktion Schaltung ", "—", 85)) + "</p>";
    content += "Relais ";
    content += IoT_WebRadioButton(" ausschalten ",      0, "BTgroup", "Roff",    0); // ID=0 ist selektiert
    content += IoT_WebRadioButton(" einschalten ",      1, "BTgroup", "Ron",     0); // ID=0 ist selektiert
    content += IoT_WebRadioButton(" invertieren ",      2, "BTgroup", "RInvert", 0); // ID=0 ist selektiert
    content += IoT_WebRadioButton(" zufällig schalten", 3, "BTgroup", "RRandom", 0); // ID=0 ist selektiert
    content += "<p>" + cleanhtml(fillcenter(" Aktion Intervall ", "—", 85)) + "</p>";
    content += "Intervall ";
    content += IoT_WebRadioButton(" aus-ein -> aus  ",  4, "BTgroup", "RIaea",   0); // ID=0 ist selektiert
    content += IoT_WebRadioButton(" aus-ein -> ein  ",  5, "BTgroup", "RIaee",   0); // ID=0 ist selektiert
    content += IoT_WebRadioButton(" ein-aus -> aus  ",  6, "BTgroup", "RIeaa",   0); // ID=0 ist selektiert
    content += IoT_WebRadioButton(" ein-aus -> ein  ",  7, "BTgroup", "RIeae",   0); // ID=0 ist selektiert
    content += "<br>Intervall ";
    content += IoT_WebRadioButton(" invertieren-invertieren -> Zustand nicht ändern ", 8, "BTgroup", "RIinv",   0); // ID=0 ist selektiert
    content += "<br><br>";
    content += IoT_WebInput("Schaltdauer: Intervall Nr. 1: ", "actA1", "01", 3);
    content += IoT_WebInput(" Minuten,  Intervall 2: ", "actA2", "01", 3);
    content += " Minuten.";
    content += "<br><br>";
    content += IoT_WebInput("Zeitpunkt Aktionsende: Stunde: ", "actAH", "00", 3);
    content += IoT_WebInput("  Minute: ", "actAM", "00", 3);
    content += "<br><br>";
    content += IoT_WebFormSubmitButton ("Timer-Aktion hinzufügen");
    content += IoT_WebFormClose();
    content += IoT_WebFormActionButton ("Abbrechen", "Abbrechen");
    content += "</body>";
    content += "</html>";
    IoT_WebServer.send(200, "text/html", content);   // HTTP-Statuscode 200 = OK (Anfrage wurde erfolgreich bearbeitet)
  }
  else
  {
    handleRoot();   
  }
}

void handleAddTimerOK ()
{
  String NewTimer = "";
  bool   Err = false;
  int    Days = 0;
  byte   Action     = 0;
  int    Stunde     = val(IoT_WebGetField("actH"));               // Stunde
  int    Minute     = val(IoT_WebGetField("actM"));               // Minute
  int    EndeStunde = val(IoT_WebGetField("actAH"));              // Stunde
  int    EndeMinute = val(IoT_WebGetField("actAM"));              // Minute
  int    Aktion1M   = val(IoT_WebGetField("actA1"));              // Aktion 1 in Minuten
  int    Aktion2M   = val(IoT_WebGetField("actA2"));              // Aktion 2 in Minuten
  String selection  = IoT_WebGetField("BTgroup");                 // Gewählter Aktionstyp
  if (selection == "Roff")
  {
    Action = 0;
    EndeStunde = 0;
    EndeMinute = 0;
    Aktion1M = 1;
    Aktion2M = 1;
  }
  else if (selection == "Ron")
  {
    Action = 1;
    EndeStunde = 0;
    EndeMinute = 0;
    Aktion1M = 1;
    Aktion2M = 1;
  }
  else if (selection == "RInvert")
  {
    Action = 2;
    EndeStunde = 0;
    EndeMinute = 0;
    Aktion1M = 1;
    Aktion2M = 1;
  }
  else if (selection == "RRandom")
  {
    Action = 3;
    EndeStunde = 0;
    EndeMinute = 0;
    Aktion1M = 1;
    Aktion2M = 1;
  }
  else if (selection == "RIaea")
  {
    Action = 8;
  }
  else if (selection == "RIaee")
  {
    Action = 9;
  }
  else if (selection == "RIeaa")
  {
    Action = 10;
  }
  else if (selection == "RIeae")
  {
    Action = 11;
  }
  else if (selection == "RIinv")
  {
    Action = 12;
  }
  else
  {
    Action = 0;                                                   // Im Fehler-Fall Relais ausschalten
    EndeStunde = 0;
    EndeMinute = 0;
    Aktion1M = 1;
    Aktion2M = 1;
  }
  listSwitch[listCount] = Action;                                 // Mit Radio-Button gewählte Aktion speichern
  if ((Stunde < 0) || (Stunde > 23))
  {
    Stunde = 99;
    Err = true;
  }
  if ((Minute < 0) || (Minute > 59))
  {
    Minute = 99;
    Err = true;
  }
  if ((EndeStunde < 0) || (EndeStunde > 23))
  {
    EndeStunde = 99;
    Err = true;
  }
  if ((EndeMinute < 0) || (EndeMinute > 59))
  {
    EndeMinute = 99;
    Err = true;
  }
  if ((Aktion1M < 1) || (Aktion1M > 240))
  {
    Aktion1M = 0;
    Err = true;
  }
  if ((Aktion2M < 1) || (Aktion2M > 240))
  {
    Aktion2M = 0;
    Err = true;
  }
  if (not(Err))                                                 // Im Falle eines Fehlers wird kein Wochentag ausgewählt
  {
    int startMinute = Stunde     * 60 + Minute;
    int endMinute   = EndeStunde * 60 + EndeMinute;
    if ((Action < 8) || (startMinute + Aktion1M + Aktion2M <= endMinute))   // Ein Intervall muss mindestens ein mal durchlaufen können
    {
      if (IoT_WebTestField("day1"))                             // Montag
      {
        Days += 2;
      }
      if (IoT_WebTestField("day2"))                             // Dienstag
      {
        Days += 4;
      }
      if (IoT_WebTestField("day3"))                             // Mittwoch
      {
        Days += 8;
      }
      if (IoT_WebTestField("day4"))                             // Donnerstag
      {
        Days += 16;
      }
      if (IoT_WebTestField("day5"))                             // Freitag
      {
        Days += 32;
      }
      if (IoT_WebTestField("day6"))                             // Samstag
      {
        Days += 64;
      }
      if (IoT_WebTestField("day7"))                             // Sonntag
      {
        Days += 1;
      }
    }
  }
  listDays  [listCount] = Days;
  listHour  [listCount] = Stunde;
  listMinute[listCount] = Minute;
  listEndHour  [listCount] = EndeStunde;
  listEndMinute[listCount] = EndeMinute;
  listAction1[listCount] = Aktion1M;
  listAction2[listCount] = Aktion2M;
  listSelected = listCount;
  listCount += 1;
  Relais_ShellSortArrays();
  handleRoot();
}

void handlePreferences ()
{
  IoT_Idle();
  String ColorHexVal     = "#000000";
  String ipaddr_web      = IoT_WLANaddress(false);
  String ipaddr_gateway  = IoT_WLANgateway(false);
  String ipaddr_subnet   = IoT_WLANsubnet(false);
  String staticInfo = " (dynamisch)";
  if (Relais_StaticIP == 1)
  {
    staticInfo = " (statisch)";
  }
  int sec = millis() / 1000;
  int min = sec / 60;
  int hr = min / 60;
  String content = StandardHeader();
  content += cleanhtml(fillcenter(" Aktuelle Netzwerk-Einstellungen ", "—", 85)) + "</p>\
    <p> </p>";
content +=
   "<p>" + cleanhtml("IP-Addresse   : " + ipaddr_web)      + "</p>\
    <p>" + cleanhtml("Gateway       : " + ipaddr_gateway)  + "</p>\
    <p>" + cleanhtml("Subnetzmaske  : " + ipaddr_subnet)   + "</p>\
    <p> </p>\
    <p>" + cleanhtml(fillcenter(" Gerätename ", "—", 85)) + "</p>\
    <p> </p>";
content +=
        IoT_WebFormOpen("NameChange") +
        IoT_WebFormSubmitButton("Ändern") +
        IoT_WebInput(" ", "devname", Relais_ID, 64) +
        IoT_WebFormClose() + "\
    <p>" + cleanhtml(fillcenter(" Netzwerk ", "—", 85)) + "</p>\
    <p> </p>";
content +=
        IoT_WebFormOpen("NetworkChange") +
      "<p>" + cleanhtml("         IP-Adresse     Gateway     Subnetzmaske") + "</p>\
      " + IoT_WebFormSubmitButton("Ändern") + "\
      " + IoT_WebInput(" ", "ip01", ipaddr_web, 15) + "\
      " + IoT_WebInput(" ", "ip02", ipaddr_gateway, 15) + "\
      " + IoT_WebInput(" ", "ip03", ipaddr_subnet, 15) + staticInfo + "\
    " + IoT_WebFormClose() + "\
    " + IoT_WebFormActionButton ("DynamicIP", "Dynamische IP-Adresse") + "\
    <p> </p>\
    <p>" + cleanhtml(fillcenter(" Sonstige ", "—", 85)) + "</p>\
    <p> </p>";
content +=
        IoT_WebFormOpen("PreferencesOK") +
        "Nach dem Neustart: " + 
        IoT_WebRadioButton(" Relais ausschalten  ", 0, "BTgroup", "Roff", Relais_BootStatus) +
        IoT_WebRadioButton(" Relais einschalten  ", 1, "BTgroup", "Ron",  Relais_BootStatus) +
        IoT_WebRadioButton(" Relais beibehalten  ", 2, "BTgroup", "Rcur", Relais_BootStatus) +
        "<p>" + IoT_WebFormSubmitButton("Einstellungen ändern") + "</p>" +
        IoT_WebFormClose();
content +=
   "<p>" + cleanhtml(fillcenter(" Einstellungen ", "—", 85)) + "</p>\
    <p> </p>\
    " + IoT_WebFormActionButton ("StandardPrefs", "Standard Einstellungen wiederherstellen") + "\
    " + IoT_WebFormActionButton ("SavePrefs", "Aktuelle Einstellungen dauerhaft speichern") + "\
    <p> </p>\
    " + IoT_WebFormActionButton ("Reload", "Startseite anzeigen")   + "\
  </body>\
</html>";

  IoT_WebUpdate(&content);                                          // HTTP-Seite aktualisieren
  content = "";
}

void handleDynamicIP ()                                             // Dynamische IP-Adresse
{
  Relais_StaticIP = 0;
  IPAddress ip1 = IPAddress(IoT_dynamicIP_address0, IoT_dynamicIP_address1, IoT_dynamicIP_address2, IoT_dynamicIP_address3);
  IPAddress ip2 = IPAddress(IoT_dynamicIP_gateway0, IoT_dynamicIP_gateway1, IoT_dynamicIP_gateway2, IoT_dynamicIP_gateway3);
  IPAddress ip3 = IPAddress(IoT_dynamicIP_subnet0,  IoT_dynamicIP_subnet1,  IoT_dynamicIP_subnet2,  IoT_dynamicIP_subnet3);
  WiFi.config(ip1, ip2, ip3);                                       // Parameter: IP, Gateway, Subnet, DNS
  handlePreferences();                                              // Einstellungen anzeigen
}

void handleNetworkChange ()                                         // Statische IP-Adresse
{
  Relais_StaticIP = 1;
  IPAddress ip1 = IPval(IoT_WebGetField("ip01"));                   // Statische IP-Adresse
  IPAddress ip2 = IPval(IoT_WebGetField("ip02"));                   // Gateway/Router IP-Adresse
  IPAddress ip3 = IPval(IoT_WebGetField("ip03"));                   // Subnet IP-Adresse
  WiFi.config(ip1, ip2, ip3);                                       // Parameter: IP, Gateway, Subnet, DNS
  handlePreferences();                                              // Einstellungen anzeigen
}

void handleNameChange ()                                            // Name ändern
{
  Relais_ID = IoT_WebGetField("devname");                            // Geräte-Name
  handlePreferences();                                              // Einstellungen anzeigen
}

void handlePreferencesOK ()                                         // Startup-Einstellungen
{
  String selection = IoT_WebGetField("BTgroup");
  if (selection == "Roff")
  {
    Relais_BootStatus = 0;
  }
  else if (selection == "Ron")
  {
    Relais_BootStatus = 1;
  }
  else if (selection == "Rcur")
  {
    Relais_BootStatus = 2;
  }
  handlePreferences();                                              // Einstellungen anzeigen
}

void handleStandardPrefs ()
{
  byte saveListCount = listCount;                                   // Aktionen sollen nicht gelöscht werden
  Relais_ParameterNew();                                            // Aktuelle Konfiguration in den Parameter-Block schreiben
  Relais_ParameterApply();                                          // Parameter-Block in die aktuelle Konfiguration übernehmen
  listCount = saveListCount;                                        // Aktionen wiederherstellen
  handlePreferences();                                              // Einstellungen anzeigen
}

void handleSavePrefs ()
{
  Relais_ParameterCreate();                                         // Aktuelle Konfiguration in den Parameter-Block schreiben
  Relais_ParameterSave();                                           // Parameter-Block in das EEPROM sichern
  IoT_EEPROMupdate();                                               // EEPROM aktualisieren
  handlePreferences();                                              // Einstellungen anzeigen
}

